from __future__ import annotations
from ehelply_bootstrapper.integrations.integration import Integration
from typing import TYPE_CHECKING
from ehelply_microservice_library.integrations.fact import get_fact_endpoint

from ehelply_bootstrapper.utils.state import State

if TYPE_CHECKING:
    from ehelply_microservice_library.routers.router_service_monitor import Heartbeat


class Monitor(Integration):
    """
    Note integration is used to talk to the ehelply-notes microservice
    """

    def __init__(self) -> None:
        super().__init__("monitor")

        self.m2m = State.integrations.get("m2m")

    def init(self):
        super().init()

    def load(self):
        super().load()

    def get_base_url(self) -> str:
        return get_fact_endpoint('ehelply-monitors')

    def register_service(self):
        """
        Register a new running service/process/thread with the monitor service
        :return:
        """
        pass

    def heartbeat(self, heartbeat: Heartbeat):
        """
        Send a heartbeat over REST
        In general, use the AWS SQS ehelply-monitoring queue to send heartbeats. This is only here in case we need
          to send one manually for some reason
        :param heartbeat:
        :return:
        """
        pass

    def minor_alarm(self):
        """
        Alert the monitoring service to a minor alarm on this microservice
        :return:
        """
        pass

    def major_alarm(self):
        """
        Alert the monitoring service to a major alarm on this microservice
        :return:
        """
        pass

    def critical_alarm(self):
        """
        Alert the monitoring service to a critical alarm on this microservice
        :return:
        """
        pass

    def warning(self):
        """
        Alert the monitoring service to a warning on this microservice
        :return:
        """
        pass

    def clear_alarm(self):
        """
        Alert the monitoring service to a clear alarm on this microservice
        :return:
        """
        pass



from ehelply_python_sdk.services.access.auth_rules import AuthRule
from ehelply_python_sdk.services.access.sdk import AuthModel


def ehelply_cloud_access(
        auth: AuthModel,
        node: str,
        service_target: str,
        exception_if_unauthorized=True
) -> bool:
    return AuthRule(
        auth,
        AuthRule(auth).participant_has_node_on_target(
            node=node,
            target_identifier=service_target,
            partition="ehelply-resources"
        ),
        AuthRule(auth).participant_has_node_on_target(
            node=node,
            target_identifier=auth.project_uuid,
            partition="ehelply-cloud"
        ).customentity_has_node_on_target(
            node=node,
            target_identifier=service_target,
            partition="ehelply-cloud",
            entity_identifier=auth.project_uuid
        ),
        exception_if_unauthorized=exception_if_unauthorized
    ).verify()



from starlette.requests import Request
from fastapi import Header, Query
from fastapi import HTTPException
from ehelply_microservice_library.utils.paginate import SQLPaginator
from ehelply_microservice_library.utils.search_filter_sort import SQLSearch, SQLSort
from ehelply_python_sdk.services.service_schemas import is_response_error


def get_fact(request: Request):
    """
    Get fact service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_fact'):
        return request.state.i_fact
    else:
        raise Exception("Fact integration has not been registered")


def get_log(request: Request):
    """
    Get log service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_log'):
        return request.state.i_log
    else:
        raise Exception("Log integration has not been registered")


def get_note(request: Request):
    """
    Get note service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_note'):
        return request.state.i_note
    else:
        raise Exception("Note integration has not been registered")


def get_meta(request: Request):
    """
    Get meta service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_meta'):
        return request.state.i_meta
    else:
        raise Exception("Meta integration has not been registered")


def get_monitor(request: Request):
    """
    Get monitor service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_monitor'):
        return request.state.i_monitor
    else:
        raise Exception("Monitor integration has not been registered")


def get_user(request: Request):
    """
    Get user service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_user'):
        return request.state.i_user
    else:
        raise Exception("User integration has not been registered")


def get_access(request: Request):
    """
    Get access service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_access'):
        return request.state.i_access
    else:
        raise Exception("Access integration has not been registered")


def get_security(request: Request):
    """
    Get security service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_security'):
        return request.state.i_security
    else:
        raise Exception("Security integration has not been registered")


def get_m2m(request: Request):
    """
    Get M2M service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_m2m'):
        return request.state.i_m2m
    else:
        raise Exception("M2M integration has not been registered")


def get_appointments(request: Request):
    """
    Get appointments service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_appointments'):
        return request.state.i_appointments
    else:
        raise Exception("Appointments integration has not been registered")


def get_places(request: Request):
    """
    Get places service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_places'):
        return request.state.i_places
    else:
        raise Exception("Places integration has not been registered")


def get_reviews(request: Request):
    """
    Get reviews service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_reviews'):
        return request.state.i_reviews
    else:
        raise Exception("Reviews integration has not been registered")


def get_notifications(request: Request):
    """
    Get notifications service integration
    :param request:
    :return:
    """
    if hasattr(request.state, 'i_notifications'):
        return request.state.i_notifications
    else:
        raise Exception("Notifications integration has not been registered")


class Integrations:
    """
    Common integrations
    """

    def __init__(
            self,
            fact,
            log,
            note,
            meta,
            monitor,
            user,
            access,
            security,
            m2m,
            appointments,
            places,
            reviews,
            notifications,
    ) -> None:
        super().__init__()

        self.fact = fact
        self.log = log
        self.note = note
        self.meta = meta
        self.monitor = monitor
        self.user = user
        self.access = access
        self.security = security
        self.m2m = m2m
        self.appointments = appointments
        self.places = places
        self.reviews = reviews
        self.notifications = notifications


def get_integrations(
        request: Request
) -> Integrations:
    """
    Dependency injection helper to get all integrations if you're lazy like me and don't want to specify a laundry list
      each time.
    Marginally lower performance on each endpoint this is used. If really trying to optimize common endpoints,
      don't use this.
    :param request:
    :return:
    """
    try:
        fact = get_fact(request=request)
    except:
        fact = None

    try:
        log = get_log(request=request)
    except:
        log = None

    try:
        note = get_note(request=request)
    except:
        note = None

    try:
        meta = get_meta(request=request)
    except:
        meta = None

    try:
        monitor = get_monitor(request=request)
    except:
        monitor = None

    try:
        user = get_user(request=request)
    except:
        user = None

    try:
        access = get_access(request=request)
    except:
        access = None

    try:
        security = get_security(request=request)
    except:
        security = None

    try:
        m2m = get_m2m(request=request)
    except:
        m2m = None

    try:
        appointments = get_appointments(request=request)
    except:
        appointments = None

    try:
        places = get_places(request=request)
    except:
        places = None

    try:
        reviews = get_reviews(request=request)
    except:
        reviews = None

    try:
        notifications = get_notifications(request=request)
    except:
        notifications = None


    return Integrations(
        fact=fact,
        log=log,
        note=note,
        meta=meta,
        monitor=monitor,
        user=user,
        access=access,
        security=security,
        m2m=m2m,
        appointments=appointments,
        places=places,
        reviews=reviews,
        notifications=notifications,
    )


def get_auth(
        request: Request,
        x_access_token: str = Header(None),
        x_secret_token: str = Header(None),
        authorization: str = Header(None),
        ehelply_active_participant: str = Header(None),
        ehelply_entity_identifier: str = Header(None),
        ehelply_project: str = Header(None),
):
    """
    Helpful dependency injection
    :param access:
    :param user:
    :param access_token:
    :param secret_token:
    :param entity_uuid:
    :param entity_type:
    :return:
    """
    from ehelply_microservice_library.integrations.access import AuthModel
    my_auth = AuthModel(access_sdk=get_access(request=request).sdk, entity_identifier=ehelply_entity_identifier, project_uuid=ehelply_project)
    m2m = get_m2m(request=request)

    if authorization:
        try:
            claims = m2m.verify_token(token=authorization)
        except:
            raise HTTPException(status_code=401, detail="Invalid token - Denied by eHelply")

        if ehelply_active_participant:
            if ehelply_active_participant in claims['custom:participants'].split(','):
                my_auth.active_participant_uuid = ehelply_active_participant
            else:
                raise HTTPException(status_code=401, detail="Invalid participant - Denied by eHelply")

        my_auth.claims = claims

    if x_secret_token and x_access_token:
        my_auth.access_token = x_access_token
        my_auth.secret_token = x_secret_token

        security = get_security(request=request)
        access = get_access(request=request)

        key = security.sdk.verify_key(access=x_access_token, secret=x_secret_token)
        if is_response_error(response=key):
            raise HTTPException(status_code=401, detail="Invalid key - Denied by eHelply")

        entity = access.sdk.get_entity_for_key(key_uuid=key.uuid)
        if is_response_error(response=entity):
            raise HTTPException(status_code=401, detail="Invalid participant for key - Denied by eHelply")

        my_auth.active_participant_uuid = entity.entity_identifier

    if not my_auth.active_participant_uuid:
        raise HTTPException(status_code=401, detail="Missing participant - Denied by eHelply")

    if not my_auth.project_uuid:
        raise HTTPException(status_code=401, detail="Missing project - Denied by eHelply")

    return my_auth


def get_pagination(
        request: Request,
        page: int = Query(1),
        page_size: int = Query(25),
):
    """
    Returns an instance of pagination which can be used with a query
    :param request:
    :param page:
    :param page_size:
    :return:
    """
    return SQLPaginator(page=page, page_size=page_size)


def get_search(
        request: Request,
        search: str = Query(None),
        search_on: str = Query(None)
):
    """
    Returns a query with a search applied.

    Args:
        request:
        search:
        search_on:

    Returns:

    """
    if search and len(search) < 3:
        raise HTTPException(status_code=400,
                            detail="Search criteria must be at least 3 characters long - Denied by eHelply")

    return SQLSearch(search_text=search, column=search_on)


def get_sort(
        request: Request,
        sort_on: str = Query(None),
        sort_desc: bool = Query(False),
):
    """
    Returns a query with a sort applied
    Args:
        request:
        sort_on:
        sort_desc:

    Returns:

    """
    return SQLSort(column=sort_on, desc=sort_desc)



from ehelply_bootstrapper.integrations.integration import Integration
from ehelply_bootstrapper.utils.state import State
from datetime import datetime
from typing import Union
from ehelply_microservice_library.integrations.fact import get_fact_endpoint


class Note(Integration):
    """
    Note integration is used to talk to the ehelply-notes microservice
    """

    def __init__(self) -> None:
        super().__init__("note")

        self.m2m = State.integrations.get("m2m")

    def init(self):
        super().init()

    def load(self):
        super().load()

    def get_base_url(self) -> str:
        return get_fact_endpoint('ehelply-notes') + "/notes"

    def create(self, content: str, author: str) -> Union[dict, bool]:
        """
        Sends a request to the notes microservice to create a new note
        :param content:
        :param author:
        :return:
        """
        if type(content) is not str or type(author) is not str:
            State.logger.warning('Note entry discarded due to invalid payload.')
            return False

        data: dict = {
            "note": {
                "content": content,
                "time": datetime.utcnow().isoformat(),
                "meta": {
                    "original_author": author,
                    "author": author,
                }
            }
        }
        return self.m2m.requests.post(self.get_base_url(), json=data).json()

    def get(self, note_uuid: str, history: int = 0, history_content: bool = True) -> Union[dict, bool, None]:
        """
        Sends a request to the notes microservice to retrieve a note
        :param note_uuid:
        :param history:
        :return:
        """
        if type(note_uuid) is not str:
            State.logger.warning('Note entry discarded due to invalid payload.')
            return False

        response = self.m2m.requests.get(self.get_base_url() + "/" + note_uuid,
                                params={"history": history, "history_content": history_content})

        if response.status_code == 404:
            return None

        return response.json()

    def delete(self, note_uuid: str, method: str = "previous") -> Union[dict, bool, None]:
        """
        Sends a request to the notes microservice to delete a note
        :param note_uuid:
        :param method:
        :return:
        """
        if type(note_uuid) is not str:
            State.logger.warning('Note entry discarded due to invalid payload.')
            return False

        response = self.m2m.requests.delete(self.get_base_url() + "/" + note_uuid, params={"method": method})

        if response.status_code == 404:
            return None

        return response.json()

    def update(self, note_uuid: str, content: str, author: str) -> Union[dict, bool, None]:
        """
        Sends a request to the notes microservice to update a note
        :param note_uuid:
        :param content:
        :param author:
        :return:
        """
        if type(content) is not str or type(author) is not str or type(note_uuid) is not str:
            State.logger.warning('Note entry discarded due to invalid payload.')
            return False

        data: dict = {
            "note": {
                "content": content,
                "time": datetime.utcnow().isoformat(),
                "meta": {
                    "author": author,
                }
            }
        }

        response = self.m2m.requests.put(self.get_base_url() + "/" + note_uuid, json=data)

        if response.status_code == 404:
            return None

        return response.json()



import typer
from ehelply_bootstrapper.utils.cryptography import Encryption, Hashing
from OpenSSL import crypto
from pathlib import Path
from typing import Optional

cli = typer.Typer()


def cert_gen(
        emailAddress="business@ehelply.com",
        commonName="microservice",
        countryName="CA",
        localityName="CA",
        stateOrProvinceName="SK",
        organizationName="eHelply",
        organizationUnitName="eHelply",
        serialNumber=0,
        validityStartInSeconds=0,
        validityEndInSeconds=10 * 365 * 24 * 60 * 60,
        KEY_FILE:Path=Path("private.key"),
        CERT_FILE:Path=Path("certificate.crt")
):
    # can look at generated file using openssl:
    # openssl x509 -inform pem -in selfsigned.crt -noout -text
    # create a key pair
    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 4096)
    # create a self-signed cert
    cert = crypto.X509()

    cert.get_subject().C = countryName
    cert.get_subject().ST = stateOrProvinceName
    cert.get_subject().L = localityName
    cert.get_subject().O = organizationName
    cert.get_subject().OU = organizationUnitName
    cert.get_subject().CN = commonName
    cert.get_subject().emailAddress = emailAddress

    cert.set_serial_number(serialNumber)

    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(validityEndInSeconds)

    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(k)

    cert.sign(k, 'sha512')

    with open(CERT_FILE, "wt") as f:
        f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))

    with open(KEY_FILE, "wt") as f:
        f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8"))


@cli.command()
def encrypt(
        data: str = typer.Option(..., prompt=True, help="Data to encrypt"),
        key: str = typer.Option(..., prompt=True, help="Encryption key. Must be a valid Fernet key")
):
    """
    Encrypt data using a fernet key
    """
    enc = Encryption([key.encode(Encryption.STRING_ENCODING)])
    typer.echo(enc.encrypt(data))


@cli.command()
def decrypt(
        data: str = typer.Option(..., prompt=True, help="Data to decrypt"),
        key: str = typer.Option(..., prompt=True, help="Encryption key. Must be a valid Fernet key")
):
    """
    Decrypt data using a fernet key
    """
    enc = Encryption([key.encode(Encryption.STRING_ENCODING)])
    typer.echo(enc.decrypt(data.encode(Encryption.STRING_ENCODING)))


@cli.command()
def hash(
        data: str = typer.Option(..., prompt=True, help="Data to hash"),
        timed: bool = typer.Option(False, help="Time how long it takes to create the hash"),
        cost: int = typer.Option(12, help="How many rounds to use to create the hash. The higher the number, the longer the hash will take")
):
    """
    Create a hash of data
    """
    Hashing.COST_FACTOR = cost
    hasher = Hashing()
    if timed:
        result = hasher.hash_timed(data)
    else:
        result = hasher.hash(data)
    typer.echo(result)


@cli.command()
def hash_check(
        data: str = typer.Option(..., prompt=True, help="Non hashed data to compare"),
        hash: str = typer.Option(..., prompt=True, help="The hash to compare against")
):
    """
    Compare data to a hash to verify whether they are equivalent
    """
    hasher = Hashing()
    typer.echo(hasher.check(data, hash))


@cli.command()
def generate_key():
    """
    Generate a new Fernet key
    """
    typer.echo(Encryption.generate_key())


@cli.command()
def refresh_https_certificates(
        key: Optional[Path] = typer.Option(None, help="Path to key file"),
        cert: Optional[Path] = typer.Option(None, help="Path to cert file")
):
    """
    DEPRECATED: Refresh HTTPS certificates
    """

    args: dict = {}

    if key:
        args['KEY_FILE'] = key
    if cert:
        args['CERT_FILE'] = cert

    cert_gen(**args)
    typer.echo("Complete")


if __name__ == '__main__':
    cli()
